# encoding: utf-8

# Exemplo de implementação de tabela hash.
#
# Página: http://malbarbo.pro.br/ensino/2013/7256/
# Autor : Marco A L Barbosa <marco@malbarbo.pro.br>

from aed import Struct, executa_testes

class Aluno(Struct):
    def __init__(self, ra, nome):
        '''
        Representa um aluno em uma universidade.
            ra: Natural - o número do registro acadêmico
            nome: String - nome do aluno
        '''
        self.init(ra, nome)


class ChaveValor(Struct):
    def __init__(self, chave, valor):
        '''
        Representa um par (chave, valor) armazenado em um tabela.
            chave: Natural
            valor: Qualquer - valor associada com a chave na tabela.
        '''
        self.init(chave, valor)


class Tabela(Struct):
    def __init__(self, n):
        '''
        Representa uma tabela hash. A tabela tem tamanho n. A função para
        calcular a posição associdada com a chave é o resto da divisão por n.
        Colições são tratadas com encadeamento.
            - n : Inteiro - tamanho da tabela
            - _valores : Lista(Lista(ChaveValor)) - representa a tabela
        '''
        self.init(n)
        self._valores = [None] * n

    def insere(self, chave, valor):
        '''
        Tabela, Inteiro, Qualquer -> None
        Adiciona a chave com o valor na tabela.
        Exemplos:
        >>> t = Tabela(5)
        >>> pedro = Aluno(54999, 'Pedro')
        >>> joao = Aluno(60001, 'Joao')
        >>> maria = Aluno(71234, 'Maria')
        >>> t.insere(54999, pedro)
        >>> t.insere(60001, joao)
        >>> t.insere(71234, maria)
        >>> t.busca(60001)
        Aluno(60001, 'Joao')
        >>> t.busca(54999)
        Aluno(54999, 'Pedro')
        >>> t.busca(71234)
        Aluno(71234, 'Maria')
        >>> t.busca(71239)
        Traceback (most recent call last):
            ...
        Exception: chave inválida
        '''
        indice = chave % self.n
        entrada = ChaveValor(chave, valor)
        if self._valores[indice] != None:
            self._valores[indice].append(entrada)
        else:
            self._valores[indice] = [entrada]
        return None

    def busca(self, chave):
        '''
        Tabela, Inteiro -> Qualquer
        Busca o valor associada com a chave na tabela.
        Lança exceção se a chave não está na tabela.
        Exemplos: Veja o método insere.
        '''
        indice = chave % self.n
        if self._valores[indice] != None:
            for entrada in self._valores[indice]:
                if entrada.chave == chave:
                    return entrada.valor
        raise Exception('chave inválida')
    
    def remove(self, chave):
        '''
        Tabela, Inteiro
        Remove uma chave e o valor associado da tabela.
        Lança exceção se a chave não está na tabela.
        Exemplos:
        >>>
        >>> t = Tabela(5)
        >>> pedro = Aluno(54999, 'Pedro')
        >>> joao = Aluno(60001, 'Joao')
        >>> maria = Aluno(71234, 'Maria')
        >>> t.insere(54999, pedro)
        >>> t.insere(60001, joao)
        >>> t.insere(71234, maria)
        >>> t.remove(54999)
        >>> t.busca(54999)  # não pode encontrar o que foi removido
        Traceback (most recent call last):
            ...
        Exception: chave inválida
        >>> t.busca(60001)  # tem que encontrar quem não foi removido
        Aluno(60001, 'Joao')
        >>> t.busca(71234)
        Aluno(71234, 'Maria')
        '''
        indice = chave % self.n
        if self._valores[indice] != None:
            for entrada in self._valores[indice]:
                if entrada.chave == chave:
                    self._valores[indice].remove(entrada)
                    return
        raise Exception('chave inválida')


executa_testes()
